#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <curses.h>

#include "wave.h"
#include "defaults.h"
#include "gsm/inc/gsm.h"
#include "dsp.h"

char *USAGE =

"USAGE gsm2wav [options] [<GSM-file>]\n"
"Converts a GSM file to a 16 bit mono wave-file or replays it on the DSP.\n"
"gsmplay is a link to gsm2wav and equivalent to 'gsm2wav -p'.\n"
"Options: -s<sample rate> in Hz, kHz or f (11025 Hz) (default 8000)\n"
"         -o<output file>, -h print this message, -v verbose mode\n"
"         -p play GSM file to DSP, -q operate quietly\n";

struct stat st;

gsm handle;
gsm_frame buf;
gsm_signal sample[160];

int main(int argc,char **argv)
{
  wavfile w,g;
  LONGINT i,j,n,s=INT_MAX;
  int c;
  VAL x;
  int replay=0,quiet=0,refresh=0,step;
  int dspfn;
  char *env;

  setexitmsg(argv[0]);
  w.f=stdout;
  w.rate=8000;
  w.bits=16;
  w.bytes=2;

  env=getenv("GSM_RATE");
  if(env && atoi(env)>0) w.rate=atoi(env);

  i=strlen(argv[0]);
  if(i>=7 && strncasecmp(argv[0]+i-7,"gsmplay",7)==0) replay=1;
  
  while(1)
  {
    c=getopt(argc,argv,"s:o:hvpq");
    if(stdopt(c,USAGE)) break;
    switch(c)
    {
      case 's':
        x=getfreq(optarg);
        if(x<0)
        {
          fprintf(stderr,"%s: Invalid sampling rate\n",PRG);
          errexit(1);
        };
        w.rate=rint(x);
        break;
      case 'o':
        w.f=fileopen(optarg,"wb",0);
        break;
      case 'p':
        replay=1;
        break;
      case 'q':
	quiet=1;
	break;
    };
  };
  
  if(argc==optind) {
    if(!replay) {
      fprintf(stderr,"%s: Can't determine size of GSM input.\n",PRG);
      errexit(ERR_FORMAT);
    };
    g.f=stdin;
    quiet=1;
  } else {

    if(optind<argc-1 || optind>=argc) errexit(ERR_ARG);

    if(stat(argv[optind],&st)) errexit(ERR_READ);
    s=st.st_size;
    if(s % 33) {
      fprintf(stderr,"%s: Input file is no GSM file.\n",PRG);
      errexit(ERR_FORMAT);
    }
    s/=33;
    g.f=fileopen(argv[optind],"r",".gsm");
  };
  if(replay) {
    dspfn=opendsp(16,w.rate,0,O_WRONLY);
  } else {
    w.n=160*s;
    writeheader(&w);
  }; 

  if (!(handle = gsm_create())) exit(1);

  if(replay && !quiet) {
    initscr();      /* initialize the curses library */
    keypad(stdscr, TRUE);  /* enable keyboard mapping */
    cbreak();       /* take input chars one at a time, no wait for \n */
    noecho();       /* don't echo input */
    nodelay(stdscr,1);
    printw("Playing %s - total time %3d min %2d sec\n\n"
      "CONTROLS:  quit 'q' or UP,     pause: SPACE or DOWN\n"
      "search backward: '<' or LEFT,  search forward: '>' or RIGHT\n"
      "skip backward: 'm' or PAGE_UP, skip forward: '/' or PAGE_UP\n\n",
      argc==optind ? "input stream" : argv[optind],
      (160*s)/(60*w.rate),((160*s)/w.rate)%60);
  }
sleep(1);
  step=5*w.rate/160;

  for(i=0;i<s && !feof(g.f);i++) {
    j=fread(buf,sizeof(buf),1,g.f);
    if(j!=1 && feof(g.f) && replay && (argc==optind)) break;
    if(j!=1 ||  gsm_decode(handle, buf, sample) < 0) {
      fprintf(stderr,"%s: Error decoding input file.\n",PRG);
      errexit(ERR_FORMAT);
    }
    if(replay) {
      write(dspfn,&sample,320);
      if(!quiet) {
	fsync(dspfn);
	c=getch();
	if(c>0) {
	  refresh=0;
	  switch(c) {
	    case 'm':
	    case 'M':
	    case KEY_PPAGE:
	      if(i>12*step) {
		if(!fseek(g.f,-12*step*33,SEEK_CUR)) i-=12*step;
		ioctl(dspfn,SNDCTL_DSP_RESET,0);
	      };
	      break;
	    case '/':
	    case '?':
	    case '-':
	    case '_':
	    case KEY_NPAGE:
	      if(i<s-12*step) {
		if(!fseek(g.f,12*step*33,SEEK_CUR)) i+=12*step;
		ioctl(dspfn,SNDCTL_DSP_RESET,0);
	      };
	      break;
	    case ',':
	    case '<':
	    case KEY_LEFT:
	      if(i>step) {
		if(!fseek(g.f,-step*33,SEEK_CUR)) i-=step;
		ioctl(dspfn,SNDCTL_DSP_RESET,0);
	      };
	      break;
	    case '.':
	    case '>':
	    case KEY_RIGHT:
	      if(i<s-step) {
		if(!fseek(g.f,step*33,SEEK_CUR)) i+=step;
		ioctl(dspfn,SNDCTL_DSP_RESET,0);
	      };
	      break;
	    case ' ':
	    case KEY_DOWN:
	      nodelay(stdscr,0);
	      getch();
	      nodelay(stdscr,1);
	      break;
	    case 'q':
	    case KEY_UP:
              ioctl(dspfn,SNDCTL_DSP_RESET,0);
	      endwin();
              exit(0);
	  };
	};

	if(refresh<=0) {
	  mvprintw(6,0,"[");
	  for(j=0;j<(57*i)/s;j++) printw("#");
          for(;j<57;j++) printw("-");
	  printw("]\n\ncurent pos: %3d min %2d sec\r",
	     (160*i)/(60*w.rate),((160*i)/w.rate)%60);
	  refresh=10;
	} else {
	  refresh--;
	};
      }
    } else {
      for(j=0;j<160;j++) {
        wavbinput(&w,sample[j]);
      }
    };
  }; 

  if(replay) {
    close(dspfn);
    if(!quiet) endwin();
  };
  errexit(0);
}
    

